#ifndef __CURSOR_H_
#define __CURSOR_H_

#include "ocicpp.h"
#include "OraError.h"
#include "OraType.h"
#include "OraString.h"
#include "OraNumber.h"
#include "OraDummy.h"
#include "OraLob.h"
#include "OraDate.h"
#include "OraRaw.h"
#include "OraLabel.h"
#include "OraRowID.h"
#include "Lob.h"
#include "RowID.h"

const string default_null_text="<NULL>";
	
class Cursor {
private: /* OCI Defined handles */
	OCIEnv *envhp;
	OCISvcCtx *svchp;
	OCIStmt *stmthp;
	OCIParam *paramd;
	OCIError *errhp;
	map<string,OCIBind *> binds; /* map of bindings */
private: 
	/* All information is in row now */
	OraType **row; /* Contains fetched row.Array of pointers to OraTypes derivative */
	sb4 nCols;
	ub2 stmtType; /* Statement type as described in table 4-1 OCI Reference */
	sword status;
	string nulltext; /* Text for null fields */
	int haveResultSet; 
	int canFetch;
	int fromBind; /* 1 if stmt was initialized from ref cursor */
	map<string,int> cols_map; /* For get<Type>("colName") methods */
public:
	Cursor();
	Cursor(OCIEnv *,OCISvcCtx *,OCIError *er);
	void init(OCIEnv *,OCISvcCtx *,OCIError *er);
	~Cursor();
	void drop();
/* Once for request*/
	int prepare(const string &);
	void exec();
	void describe();
	void define();
	void hashHead(); 
	void execute(); /* User's should use this func in most cases 
	                   or they can do exec(),describe(),define()
	                   by hand
	                */
	void bind(const string &par,const string &val);
	void bind(const string &par,int val);
	void bind(const string &par,double val);
	void bind(const string &par,RowID &val);
	void bind(const string &par,Cursor &val);
	
/* fetch returns non zero value if theres another rows */
	int fetch();
/* get<Type> methods working with hardcoded cols numbers */	
	void getStr(int col,string &) const;
	void getInt(int col,int &val) const;
	void getDouble(int col,double &val) const;
	void getStrMon(int col,string &val) const;
	void getSec(int col,int &val) const;
	void getMin(int col,int &val) const;
	void getHour(int col,int &val) const;
	void getDay(int col,int &val) const;
	void getMonth(int col,int &val) const;
	void getYear(int col,int &val) const;
	void getCLOB(int col,Lob &) const;
	void getBLOB(int col,Lob &) const;
	void getRowID(int col,RowID &) const;
	
/* get<Type> same but working with attribute name */
	void getStr(const string &col,string &val) const;
	void getInt(const string &col,int &val) const;
	void getDouble(const string &col,double &val) const;
	void getStrMon(const string &,string &) const;
	void getSec(const string &,int &) const;
	void getMin(const string &,int &) const;
	void getHour(const string &,int &) const;
	void getDay(const string &,int &) const;
	void getMonth(const string &,int &) const;
	void getYear(const string &,int &) const;
	void getCLOB(const string &col,Lob &lob) const;
	void getBLOB(const string &col,Lob &lob) const;
	void getRowID(const string &col,RowID &rid) const;
	
/* Some functions to describe select-list */	
	int getNCols() const;
	void getColName(int col,string &) const; 
	int getColSize(int col) const;
	int getColSize(const string &) const;
	int getColType(int col) const;
	int getColType(const string &) const;
	void getColTypeName(int col,string &) const;
	void getColTypeName(const string &,string &) const;
	
/* Some supplementary functions */
	void setNullText(const string &);
private:			
	void newCellByType(OraType **,OCIStmt *,int col);

};
#endif